#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
loff_t lseek64 (int, loff_t, int);

#define MYNAME "length"

int shorter (int fd, long long pos)
{
	int i;
	char dummy[1];

	pos--;
	if (pos < 0) return 0;
	if (lseek64(fd, pos, SEEK_SET) < 0) {
		fprintf(stderr, MYNAME ": lseek64: %s\n", sys_errlist[errno]);
		exit(1);
	}
	i = read(fd, dummy, 1);
	if (i < 0) {
		fprintf(stderr, MYNAME ": read: %s\n", sys_errlist[errno]);
		exit(1);
	}
	return !i;
}

int main (int argc, char *argv[])
{
	int fd;
	struct stat st;
	long long pos = 0x1000000;
	long long step;

	if (argc != 2 || !strcmp(argv[1], "--help")) {
		fprintf(stderr, "Usage: " MYNAME " device\n"
		  "For example: " MYNAME " /dev/sda1\n"
		  "or:          " MYNAME " /etc/passwd\n");
		exit(1);
	}

	if (stat(argv[1], &st) < 0) {
		fprintf(stderr, MYNAME ": stat: %s\n", sys_errlist[errno]);
		exit(1);
	}

	if (S_ISREG(st.st_mode)) {
		printf("%ld\n", st.st_size);
		return 0;
	}

	if (S_ISBLK(st.st_mode)) {
		if ((fd = open(argv[1], O_RDONLY)) < 0) {
			fprintf(stderr, MYNAME ": open: %s\n", sys_errlist[errno]);
			exit(1);
		}
		while (!shorter(fd, pos)) pos *= 2;
		pos /= 2;
		step = pos/2;
		while (step) {
			if (shorter(fd, pos)) pos -= step;
			else pos += step;
			step /= 2;
		}
		if (shorter(fd, pos)) pos--;
		close(fd);
		printf("%lld\n", pos);
		return 0;
	}

	fprintf(stderr, MYNAME ": file is neither block device nor regular file\n");
	return 1;
}
